//-*-c++-*------------------------------------------------------------
//
// File name : bioArithMult.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Apr 28 20:44:04 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"

#include "bioArithMult.h"
#include "bioArithBinaryPlus.h"
#include "bioArithConstant.h"
#include "bioArithNamedExpression.h"
#include "patError.h"
#include "bioLiteralRepository.h"
#include "bioExpressionRepository.h"
#include "bioArithCompositeLiteral.h"

bioArithMult::bioArithMult(bioExpressionRepository* rep,
			   patULong par,
			   patULong left, 
			   patULong right,
			   patError*& err) : 
  bioArithBinaryExpression(rep, par,left,right,err)
{
}

bioArithMult::~bioArithMult() {}

patString bioArithMult::getOperatorName() const {
  return ("*") ;
}

patReal bioArithMult::getValue(patBoolean prepareGradient, patError*& err) {
  
  patReal leftVal = leftChild->getValue(prepareGradient,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  if (leftVal == 0.0) {
    return 0.0 ;
  }

  patReal rightVal = rightChild->getValue(prepareGradient,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  if (rightVal == 0.0) {
    return 0.0 ;
  }

  patReal result = leftVal * rightVal ;

  return result ;
}

bioExpression* bioArithMult::getDerivative(patULong aLiteralId, 
					   patError*& err) const {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }

  
  bioExpression* leftValue = leftChild->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* rightValue = rightChild->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  patBoolean leftOne(patFALSE) ;
  bioExpression* leftDeriv = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  if (leftDeriv->isConstant()) {
    patReal v = leftDeriv->getValue(patFALSE,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL;
    }
    if (v == 1) {
      leftOne = patTRUE ;
    }
  }
  patBoolean rightOne(patFALSE) ;
  bioExpression* rightDeriv = rightChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  if (rightDeriv->isConstant()) {
    patReal v = rightDeriv->getValue(patFALSE,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL;
    }
    if (v == 1) {
      rightOne = patTRUE ;
    }
  }

  

  bioExpression* l ;
  if (leftOne) {
    l = rightValue->getDeepCopy(theRepository,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL;
    }
  }
  else {
    l = new bioArithMult(theRepository,patBadId,leftDeriv->getId(),rightValue->getId(),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL;
    }
  }
  bioExpression* r ;
  if (rightOne) {
    r = leftValue->getDeepCopy(theRepository,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL;
    }
  }
  else {
    r = new bioArithMult(theRepository,patBadId,leftValue->getId(),rightDeriv->getId(),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL;
    }
  }
  if (leftChild->dependsOf(aLiteralId)) {
    if (rightChild->dependsOf(aLiteralId)) {
      // Left variable, right variable
      bioExpression* result = new bioArithBinaryPlus(theRepository,patBadId,l->getId(),r->getId(),err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL;
      }
      return result ;
    }
    else {
      // Left variable, right constant
      return l ;
    }
  }
  else {
    if (rightChild->dependsOf(aLiteralId)) {
      //Left constant, right variable
      return r ;
    }
    else {
      //Left constant, right constant
      bioExpression* result = theRepository->getZero() ;
      return result ;
    }
  }


}

bioArithMult* bioArithMult::getDeepCopy(bioExpressionRepository* rep,
					patError*& err) const {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithMult* theNode = 
    new bioArithMult(rep,patBadId,leftClone->getId(),rightClone->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  

  return theNode ;
}

bioArithMult* bioArithMult::getShallowCopy(bioExpressionRepository* rep,
					patError*& err) const {
  bioArithMult* theNode = 
    new bioArithMult(rep,patBadId,leftChild->getId(),rightChild->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  

  return theNode ;
}

patString bioArithMult::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patBoolean lSimple = leftChild->isSimple() ;
  patString vLeft ;
  if (lSimple) {
    vLeft = leftChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vLeft = leftChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  patBoolean rSimple = rightChild->isSimple() ;
  patString vRight ;
  if (rSimple) {
    vRight = rightChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vRight = rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  stringstream str ;
  str << getCppResultDeclaration(prefix) << ";" << endl ;
  if (!lSimple) {
    str << vLeft << endl ;
  }
  if (!rSimple) {
    str << vRight << endl ;
  }
  if (!lSimple) {
    str << "if (" <<leftChild->getCppResult(prefix) << "== 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  else {
    str << "if ((" << vLeft << ") == 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  if (!rSimple) {
    str << "else if (" <<rightChild->getCppResult(prefix) << "== 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  else {
    str << "else if ((" << vRight << ") == 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  str << " else {" << endl ;
  str << getCppResult(prefix) << "= (" ;
  if (lSimple) {
    str << vLeft ;
  }
  else {
    str << leftChild->getCppResult(prefix) ;
  }
  str << " ) * ( " ;
  if (rSimple) {
    str << vRight ;
  }
  else {
    str << rightChild->getCppResult(prefix) ;
  }
  str << ");" << endl ;
  str << "    if (!finite(" << getCppResult(prefix) << ")) {" << endl ;
  str << "      " << getCppResult(prefix) << " =  patMaxReal ;" << endl ;
  str << "    }" << endl ;

  str << "}" << endl ;
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }

}


patString bioArithMult::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "(" << vLeft << ")*(" << vRight << ")"  ;
  return patString(str.str()) ;
}

patBoolean bioArithMult::isStructurallyZero() const {
  if (leftChild->isStructurallyZero()) {
    return patTRUE ;
  }
  if (rightChild->isStructurallyZero()) {
    return patTRUE ;
  }
  return patFALSE ;
}

patBoolean bioArithMult::isSimple() const {
  return patFALSE ;
}

patString bioArithMult::getExpressionString() const {
  stringstream str ;
  str << '*' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}






bioFunctionAndDerivatives* bioArithMult::getNumericalFunctionAndGradient(vector<patULong> literalIds,
									patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }
  bioFunctionAndDerivatives* l = leftChild->getNumericalFunctionAndGradient(literalIds,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  if (l->theFunction == 0) {
    // l = 0
    patReal r = rightChild->getValue(patFALSE,err) ; 
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (r == 0) {
      // l = 0, r = 0
      result.theFunction =  0.0 ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	result.theGradient[i] = 0.0 ;
      }
    }
    else {
      // l = 0, r != 0
      result.theFunction =  0.0 ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	if (l->theGradient[i] == 0.0) {
	  result.theGradient[i] = 0.0 ;
	}
	else {
	  result.theGradient[i] = l->theGradient[i] * r ;
	}
      }
    }
  }
  else {
    // l != 0 
    bioFunctionAndDerivatives* r = rightChild->getNumericalFunctionAndGradient(literalIds,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (r->theFunction == 0) {
      // l != 0, r = 0
      result.theFunction = 0.0 ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	if (r->theGradient[i] == 0.0) {
	  result.theGradient[i] = 0.0 ;
	}
	else {
	  result.theGradient[i] = r->theGradient[i] * l->theFunction ;
	}
      }
    }
    else {
      // l != 0, r = 0
      result.theFunction =  l->theFunction * r->theFunction ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	result.theGradient[i] = l->theGradient[i] * r->theFunction + 
	  r->theGradient[i] * l->theFunction ;
      }
      
    }
  }

  if (performCheck) {
    for (patULong i = 0 ; i < literalIds.size() ; ++i) {
      if (!finite(result.theGradient[i])) {
	result.theGradient[i] = patMaxReal ;
      }
    }
  }
  return &result ;
}

