from biogeme import *

## Computes the null loglikelihood from the sample and ask Biogeme to include it in the output file.
# \param availability A <a href="http://docs.python.org/py3k/tutorial/datastructures.html#dictionaries" target="_blank">dictionary</a> mapping each alternative ID with its availability condition.
# \param iterator An iterator on the data file.
# \return loglikelihood of a model where the choice probability for
# observation \f$n\f$ is given by is \f$1/J_n\f$, where \f$J_n\f$ is
# the number of available alternatives, i.e. \f[ \mathcal{L} = -\sum_n \ln(J_n) \f]
def nullLoglikelihood(availability,iterator):
    total = 0 
    for i,a in availability.items() :
        total += (a != 0)
    nl = -Sum(log(total),iterator)
    BIOGEME_OBJECT.STATISTICS['Null loglikelihood'] = nl
    return nl

## Computes the number of times each alternative is chosen in the data set and ask Biogeme to include it in the output file..
# \param choiceSet list containing the alternatives for which statistics must be computed.
# \param choice expression producing the id of the chosen alternative.
# \param iterator An iterator on the data file.
# \return A <a href="http://docs.python.org/py3k/tutorial/datastructures.html#dictionaries" target="_blank">dictionary</a> n with an entry n[i] for each alternative i containing the number of times it is chosen.
def choiceStatistics(choiceSet,choice,iterator):
    # Note that availability is ignored here
    n = {}
    for i in choiceSet:
        n[i] = Sum(choice == i,iterator)
    for i in choiceSet:
        s = 'Alt. %d chosen' % (i)
        BIOGEME_OBJECT.STATISTICS[s] = n[i]
    return n

## Computes the number of times each alternative is declared available in the data set and ask Biogeme to include it in the output file..
# \param availability <a href="http://docs.python.org/py3k/tutorial/datastructures.html#dictionaries" target="_blank">Dictionary</a> containing for each alternative the expression for its availability. 
# \param iterator An iterator on the data file.
# \return A <a href="http://docs.python.org/py3k/tutorial/datastructures.html#dictionaries" target="_blank">dictionary</a> n with an entry n[i] for each alternative i containing the number of times it is available.
def availabilityStatistics(availability,iterator):
    n = {}
    for i,a in availability.items():
        n[i] = Sum((a != 0),iterator)
    for i,a in availability.items():
        s = 'Alt. %d available' % (i)
        BIOGEME_OBJECT.STATISTICS[s] = n[i]
    return n    

## Computes the constant loglikelihood from the sample and ask Biogeme to include it in the output file. It assumes that the full choice set is available for each observation.
# \param availability A <a href="http://docs.python.org/py3k/tutorial/datastructures.html#dictionaries" target="_blank">dictionary</a> mapping each alternative ID with its availability condition.
# \param choiceSet list containing the alternatives in the choice set.
# \param choice expression producing the id of the chosen alternative.
# \param iterator An iterator on the data file.
# \return loglikelihood of a logit model where the only parameters are the alternative specific constant. If \f$n_i\f$ is the number of time alternative \f$i\f$ is chosen, then it is given by \f[ \mathcal{L} = \sum_i n_i \ln(n_i) - n \ln(n)  \f] where \f$ n = \sum_i n_i \f$ is the total number of observation.
def cteLoglikelihood(choiceSet,choice,iterator):
    n = choiceStatistics(choiceSet,choice,iterator)
    l = 0
    total = 0
    for i in n:
        l += n[i] * log(n[i])
        total += n[i]
    l -= total * log(total)
    BIOGEME_OBJECT.STATISTICS['Cte loglikelihood (only for full choice sets)'] = l
    return l

